// funs.cpp:  supporting functions for the MCMC
//            - suff stats for gibbs draws
//            - suff stats for MH steps
//            - serial and parallel codepaths
//--------------------------------------------------
#include <cmath>
#include <algorithm> //for binary_search
#include "funs.h"
#include <map>
#ifdef CALIBART
#include "calibd.h" //for minperbot
#endif

//--------------------------------------------------
// Evaluate normal density at x given mean m, variance v 
double pn(double x, double m, double v)
{
	double dif = x-m;
	return exp(-.5*dif*dif/v)/sqrt(2*PI*v);
}

//--------------------------------------------------
// Draw from discrete distribution given by p using random
// number generator gen
// Return index
int rdisc(double *p, RNG& gen)
{
	double sum;
	double u = gen.uniform();
	int i=0;

	sum=p[0];
	while(sum<u) {
		i += 1;
		sum += p[i];
	}
	return i;
}

//--------------------------------------------------
// Evaluate tree tr on grid given by cutpoint info xi and write to os
void grm(tree& tr, xinfo& xi, std::ostream& os) 
{
	size_t p = xi.size();
	if(p!=2) {
		cout << "error in grm, p !=2\n";
		return;
	}
	size_t n1 = xi[0].size();
	size_t n2 = xi[1].size();
	tree::tree_cp bp; //pointer to bottom node
	double *x = new double[2];

	for(size_t i=0;i!=n1;i++) {
		for(size_t j=0;j!=n2;j++) {
			x[0] = xi[0][i]; 
			x[1] = xi[1][j]; 
			bp = tr.bn(x,xi);
			os << x[0] << " " << x[1] << " " << bp->getm() << " " << bp->nid() << endl;
		}
	}
	delete[] x;
}

//--------------------------------------------------
// Does this bottom node n have any variables it can split on.
bool cansplit(tree::tree_p n, xinfo& xi)
{
	int L,U;
	bool v_found = false; //have you found a variable you can split on
	size_t v=0;

	while(!v_found && (v < xi.size())) { //invar: splitvar not found, vars left
		L=0; U = xi[v].size()-1;
		n->rg(v,&L,&U);
		if(U>=L) v_found=true;
		v++;
	}
	return v_found;
}

//--------------------------------------------------
// Compute prob of a birth for tree t with cutpoints xi and prior pi, 
// goodbots will contain all the good bottom nodes
double getpb(tree& t, xinfo& xi, pinfo& pi, tree::npv& goodbots)
{
	double pb;  //prob of birth to be returned
	tree::npv bnv; //all the bottom nodes

	t.getbots(bnv);
	for(size_t i=0;i!=bnv.size();i++) 
		if(cansplit(bnv[i],xi)) goodbots.push_back(bnv[i]);
	if(goodbots.size()==0) { //are there any bottom nodes you can split on?
		pb=0.0;
	} else { 
		if(t.treesize()==1) pb=1.0; //is there just one node?
		else pb=pi.pb;
	}
	return pb;
}
//--------------------------------------------------
// Find variables tree n can split on, put their indices in goodvars
void getgoodvars(tree::tree_p n, xinfo& xi,  std::vector<size_t>& goodvars)
{
	int L,U;

	for(size_t v=0;v!=xi.size()-1;v++) {//try each variable except Im which is the last one. If we want to try every var, should just be xi.size().
		L=0; U = xi[v].size()-1;
		n->rg(v,&L,&U);
		if(U>=L) goodvars.push_back(v);
	}
}

bool isImgoodvar(tree::tree_p n, xinfo& xi)
{
	int L,U;
	size_t Imvar=xi.size();

	L=0; U=xi[Imvar].size()-1;
	n->rg(Imvar,&L,&U);
	if(U>=L) return true;

	return false;
}

//--------------------------------------------------
// Get prob a node n grows, 0 if no good vars, else alpha/(1+d)^beta
// with prior depth penalty params alpha, beta 
double pgrow(tree::tree_p n, xinfo& xi, pinfo& pi)
{
	if(cansplit(n,xi)) {
		return pi.alpha/pow(1.0+n->depth(),pi.beta);
	} else {
		return 0.0;
	}
}

//--------------------------------------------------
// Get the L,U values for a node in the tree *given* the tree
// structure both above and below that node.
void getLU(tree::tree_p pertnode, xinfo& xi, int* L, int* U)
{
	tree::tree_p l,r;

	*L=0; *U = xi[pertnode->getv()].size()-1;
	l=pertnode->getl();
	r=pertnode->getr();

	bool usel,user;
	usel=l->nuse(pertnode->getv());
	user=r->nuse(pertnode->getv());
	if(usel && user)
	{
		l->rl(pertnode->getv(),L);
		r->ru(pertnode->getv(),U);
	}
	else if(usel)
	{
		pertnode->rg(pertnode->getv(),L,U);
		l->rl(pertnode->getv(),L);
	}
	else
	{
		pertnode->rg(pertnode->getv(),L,U);
		r->ru(pertnode->getv(),U);
	}
}

//--------------------------------------------------
// similar except we get it for a prescribed variable pertvar
void getpertLU(tree::tree_p pertnode, size_t pertvar, xinfo& xi, int* L, int* U)
{
		*L=0; *U = xi[pertvar].size()-1;

		bool usel,user;
		usel=pertnode->l->nuse(pertvar);
		user=pertnode->r->nuse(pertvar);
		if(usel && user)
		{
			pertnode->l->rl(pertvar,L);
			pertnode->r->ru(pertvar,U);
		}
		else if(usel)
		{
			pertnode->rg(pertvar,L,U);
			pertnode->l->rl(pertvar,L);
		}
		else
		{
			pertnode->rg(pertvar,L,U);
			pertnode->r->ru(pertvar,U);
		}
}

//--------------------------------------------------
// number of available cutpoints at node n for variable var
int getnumcuts(tree::tree_p n, xinfo& xi, size_t var)
{
	int L,U;

	getpertLU(n,var,xi,&L,&U);
	return std::max(0,U-L+1);
}

//--------------------------------------------------
// Find numbr of variables internal tree node n can split on
void getinternalvars(tree::tree_p n, xinfo& xi,  std::vector<size_t>& goodvars)
{
   int L,U;

   for(size_t v=0;v!=xi.size();v++) {//try each variable
      L=0; U = xi[v].size()-1;
      getpertLU(n,v,xi,&L,&U);
      if(U>=L) goodvars.push_back(v);
   }
}


//--------------------------------------------------
// Get sufficients stats for *all* bottom nodes of tree x.
// Covariate info given by di and suff stats sv are updated
void allsuff(tree& x, xinfo& xi, dinfo& di, tree::npv& bnv, std::vector<sinfo>& sv)
{
	tree::tree_cp tbn; //the pointer to the bottom node for the current observations
	size_t ni;         //the  index into vector of the current bottom node
	double *xx;        //current x
	double y;          //current y

	bnv.clear();
	x.getbots(bnv);

	typedef tree::npv::size_type bvsz;
	bvsz nb = bnv.size();
	sv.resize(nb);

	std::map<tree::tree_cp,size_t> bnmap;
	for(bvsz i=0;i!=bnv.size();i++) bnmap[bnv[i]]=i;

	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		y=di.y[i];

		tbn = x.bn(xx,xi);
		ni = bnmap[tbn];

		++(sv[ni].n);
		sv[ni].sy += y;
		sv[ni].sy2 += y*y;
	}
}

//--------------------------------------------------
// Get sufficient stats for children (v,c) of node nx in tree x
// Covariate info in di
// Updates suff stats for left and right children in sl, sr
void getsuff(tree& x, tree::tree_cp nx, size_t v, size_t c, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr)
{
	double *xx; //current x
	double y;   //current y

	sl.n=0;sl.sy=0.0;sl.sy2=0.0;
	sr.n=0;sr.sy=0.0;sr.sy2=0.0;

	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		if(nx==x.bn(xx,xi)) {     //does the bottom node = xx's bottom node
			y = di.y[i];
			if(xx[v] < xi[v][c]) { //left child
				sl.n++;
				sl.sy += y;
				sl.sy2 += y*y;
			} else {              //right child
				sr.n++;
				sr.sy += y;
				sr.sy2 += y*y;
			}
		}
	}
}

//--------------------------------------------------
// similar, but in case of perturb.
void getpertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, xinfo& xi, dinfo& di, std::vector<sinfo>& svold, std::vector<sinfo>& svnew)
{
	allsuff(x,xi,di,bnv,svnew);
	pertnode->setc(oldc);
	allsuff(x,xi,di,bnv,svold);		
}
void getpertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, size_t oldvar, bool didswaplr, xinfo& xi, dinfo& di, std::vector<sinfo>& svold, std::vector<sinfo>& svnew)
{
	allsuff(x,xi,di,bnv,svnew);
	pertnode->setv(oldvar);
	pertnode->setc(oldc);
	if(didswaplr) //undo swap
	{
		tree::tree_p tempt;
		tempt=pertnode->r;
		pertnode->r=pertnode->l;
		pertnode->l=tempt;
	}
	allsuff(x,xi,di,bnv,svold);		
}


//--------------------------------------------------
// Get sufficient stats for pair of bottom children nl(left) and nr(right) in tree x
// Covariate info in di
// Updates suff stats for left and right children in sl, sr
void getsuff(tree& x, tree::tree_cp nl, tree::tree_cp nr, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr)
{
	double *xx; //current x
	double y;   //current y

	sl.n=0;sl.sy=0.0;sl.sy2=0.0;
	sr.n=0;sr.sy=0.0;sr.sy2=0.0;

	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		tree::tree_cp bn = x.bn(xx,xi);
		if(bn==nl) { //left child
			y = di.y[i];
			sl.n++;
			sl.sy += y;
			sl.sy2 += y*y;
		}
		if(bn==nr) { //right child
			y = di.y[i];
			sr.n++;
			sr.sy += y;
			sr.sy2 += y*y;
		}
	}
}

//--------------------------------------------------
// Log of the integrated likelihood - see CGM2010 for example.
// Is a function of n, sum sy, sum of squares sy2, sigma and tau
double lil(size_t n, double sy, double sy2, double sigma, double tau)
{
	double yb,yb2,S,sig2,d;
	double sum, rv;

	yb = sy/n;
	yb2 = yb*yb;
	S = sy2 - (n*yb2);
	sig2 = sigma*sigma;
	d = n*tau*tau + sig2;
	sum = S/sig2 + (n*yb2)/d;
	rv = -(n*LTPI/2.0) - (n-1)*log(sigma) -log(d)/2.0;
	rv = rv -sum/2.0;
	return rv;
}

//--------------------------------------------------
//log of the integrated likelihood for the calibart model.  
//we dispense with terms that cancel out in the log ratio.
double clil(size_t n, size_t nf, double sy, double sf, double sigma, double sigmaf, double t2, double mu0)
{
	double C=((int)n)/sigma/sigma+((int)nf)/sigmaf/sigmaf;
	double T=1.0/t2;
	double CpT=C+T;
	double BpTmu=sy/sigma/sigma+sf/sigmaf/sigmaf+mu0/t2;
	return -.5*log(t2*C+1.0)+.5*BpTmu*BpTmu/CpT-.5*mu0*mu0/t2;
}

//--------------------------------------------------
// Get them all for b/d using clil
void calc_all_lil(tree::tree_p nx, size_t v, double* lill, double* lilr, double* lilt, sinfo& sl, sinfo& sr, pinfo& pi)
{
			if(nx->isrightchildofvsplit(pi.imdim))   //use prior on mu for bias case.  Here, sl.nf must be >=minperbot
				*lill = clil(0,sl.nf,0,sl.sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
			else if(nx->isleftchildofvsplit(pi.imdim) || v==pi.imdim)	//use comp model prior on mu. Here, sl.nf=0, sl.n must be >=minperbot
				*lill = clil(sl.n,0,sl.sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
			else  // use mu0,tau0^2 prior
				*lill = clil(sl.n,sl.nf,sl.sy,sl.sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);

			if(nx->isrightchildofvsplit(pi.imdim) || v==pi.imdim) //use prior on mu for bias case. Here, sr.nf must be >=minperbot
				*lilr = clil(0,sr.nf,0.0,sr.sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
			else if(nx->isleftchildofvsplit(pi.imdim)) //use comp model prior on mu.  Here, sl.nf=0, sl.n must be >=minperbot
				*lilr = clil(sr.n,0,sr.sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
			else // use m0,tau0^2 prior
				*lilr = clil(sr.n,sr.nf,sr.sy,sr.sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);

			if(nx->isrightchildofvsplit(pi.imdim)) //use prior on mu for bias case.  Here, sl.nf and sr.nf must be >=minperbot
				*lilt = clil(0,sl.nf+sr.nf,0.0,sl.sf+sr.sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
			else if(nx->isleftchildofvsplit(pi.imdim))  //use comp model prior on mu
				*lilt = clil(sl.n+sr.n,0,sl.sy+sr.sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
			else
				*lilt = clil(sl.n+sr.n,sl.nf+sr.nf,sl.sy+sr.sy,sl.sf+sr.sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
}


#ifdef _OPENMP
//--------------------------------------------------
// fit
void fit(tree& t, xinfo& xi, dinfo& di, std::vector<double>& fv)
{
	double *xx;
	tree::tree_cp bn;

	fv.resize(di.n);
#pragma omp parallel for private(xx,bn)
	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		bn = t.bn(xx,xi);
		fv[i] = bn->getm();
   }
}

//--------------------------------------------------
// fit
void fit(tree& t, xinfo& xi, dinfo& di, double* fv)
{
	double *xx;
	tree::tree_cp bn;

#pragma omp parallel for private(xx,bn)
	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		bn = t.bn(xx,xi);
		fv[i] = bn->getm();
   }
}

#else 
//_OPENMP
//--------------------------------------------------
// fit
void fit(tree& t, xinfo& xi, dinfo& di, std::vector<double>& fv)
{
	double *xx;
	tree::tree_cp bn;

	fv.resize(di.n);
	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		bn = t.bn(xx,xi);
		fv[i] = bn->getm();
   }
}

//--------------------------------------------------
// fit
void fit(tree& t, xinfo& xi, dinfo& di, double* fv)
{
	double *xx;
	tree::tree_cp bn;

	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		bn = t.bn(xx,xi);
		fv[i] = bn->getm();
   }
}
#endif 
//_OPENMP

//--------------------------------------------------
// fit at one x input
double fitone(tree& t,xinfo& xi, dinfo& di, size_t ix)
{
	double *xx;
	tree::tree_cp bn;
	xx = di.x + ix*di.p;
	bn = t.bn(xx,xi);
	return bn->getm();
}

//--------------------------------------------------
// partition
void partition(tree& t, xinfo& xi, dinfo& di, std::vector<size_t>& pv)
{
	double *xx;
	tree::tree_cp bn;

	pv.resize(di.n);
	for(size_t i=0;i<di.n;i++) {
		xx = di.x + i*di.p;
		bn = t.bn(xx,xi);
		pv[i] = bn->nid();
	}
}

//--------------------------------------------------
// draw all the bottom node mu's for tree t using covariate info di, priors pi
// and random number generator gen
void drmu(tree& t, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen)
{
	tree::npv bnv;
	std::vector<sinfo> sv;

	allsuff(t,xi,di,bnv,sv);

	double a = 1.0/(pi.tau0 * pi.tau0);
	double sig2 = pi.sigma * pi.sigma;
	double b,ybar;

	for(tree::npv::size_type i=0;i!=bnv.size();i++) {
		b = sv[i].n/sig2;
		ybar = sv[i].sy/sv[i].n;
		bnv[i]->setm(b*ybar/(a+b) + gen.normal()/sqrt(a+b));
	}
}

//--------------------------------------------------
// Write cutpoint information to screen
void prxi(xinfo& xi)
{
	cout << "xinfo: \n";
	for(size_t v=0;v!=xi.size();v++) {
		cout << "v: " << v << endl;
		for(size_t j=0;j!=xi[v].size();j++) cout << "j,xi[v][j]: " << j << ", " << xi[v][j] << endl;
	}
	cout << "\n\n";
}

//--------------------------------------------------
// Make xinfo = cutpoints
void makexinfo(size_t p, size_t n, double *x, xinfo& xi, size_t nc)
{
	double xinc;

	//compute min and max for each x
	std::vector<double> minx(p,INFINITY);
	std::vector<double> maxx(p,-INFINITY);
	double xx;

	for(size_t i=0;i<p;i++) {
		for(size_t j=0;j<n;j++) {
			xx = *(x+p*j+i);
			if(xx < minx[i]) minx[i]=xx;
			if(xx > maxx[i]) maxx[i]=xx;
		}
	}
	//make grid of nc cutpoints between min and max for each x.
	xi.resize(p);
	for(size_t i=0;i<p;i++) {
		xinc = (maxx[i]-minx[i])/(nc+1.0);
		xi[i].resize(nc);
		for(size_t j=0;j<nc;j++) xi[i][j] = minx[i] + (j+1)*xinc;
	}
}

//--------------------------------------------------
// Get min/max needed to make cutpoints for the p covariates
void makeminmax(size_t p, size_t n, double *x, std::vector<double> &minx, std::vector<double> &maxx)
{
	double xx;

	for(size_t i=0;i<p;i++) {
		for(size_t j=0;j<n;j++) {
			xx = *(x+p*j+i);
			if(xx < minx[i]) minx[i]=xx;
			if(xx > maxx[i]) maxx[i]=xx;
		}
	}
}

//--------------------------------------------------
// Make xinfo = cutpoints given the minx and maxx vectors and number of cutpoints nc
void makexinfominmax(size_t p, xinfo& xi, size_t nc, std::vector<double> &minx, std::vector<double> &maxx)
{
	double xinc;

	//make grid of nc cutpoints between min and max for each x.
	xi.resize(p);
	for(size_t i=0;i<p;i++) {
		xinc = (maxx[i]-minx[i])/(nc+1.0);
		xi[i].resize(nc);
		for(size_t j=0;j<nc;j++) xi[i][j] = minx[i] + (j+1)*xinc;
	}
}

//--------------------------------------------------
// Maxe xinfo = cutpoints given the minx and maxx vectors and a vector nc representing number of cutpoints for each dimension
void makexinfominmax(size_t p, xinfo& xi, std::vector<size_t> &nc, std::vector<double> &minx, std::vector<double> &maxx)
{
	double xinc;

	//make grid of nc cutpoints between min and max for each x.
	xi.resize(p);
	for(size_t i=0;i<p;i++) {
		xinc = (maxx[i]-minx[i])/(nc[i]+1.0);
		xi[i].resize(nc[i]);
		for(size_t j=0;j<nc[i];j++) xi[i][j] = minx[i] + (j+1)*xinc;
	}
}

#ifdef CALIBART

#ifdef _OPENMP
//--------------------------------------------------
// Get sufficient stats for children (v,c) of node nx in tree x for the CALIBART calibration model
void getCalisuff(tree& x, tree::tree_cp nx, size_t v, size_t c, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr)
{
   double *xx;//current x
   double y;  //current y
   sl.n=0;sl.nf=0;sl.sy=0.0;sl.sf=0.0;sl.sy2=0.0;sl.sf2=0.0;
   sr.n=0;sr.nf=0;sr.sy=0.0;sr.sf=0.0;sr.sy2=0.0;sr.sf2=0.0;

   size_t ln,rn;
   double lsy,lsy2,rsy,rsy2;
   size_t lnf,rnf;
   double lsf,lsf2,rsf,rsf2;

#pragma omp parallel private(xx,y,ln,rn,lsy,rsy,lsy2,rsy2)
{
   ln=0;rn=0;lsy=0.0,rsy=0.0,lsy2=0.0,rsy2=0.0;
#pragma omp for nowait
   for(size_t i=0;i<di.n;i++) {
      xx = di.x + i*di.p;
      if(nx==x.bn(xx,xi)) { //does the bottom node = xx's bottom node
         y = di.y[i];
         if(xx[v] < xi[v][c]) {
		ln++;
		lsy += y;
		lsy2 += y*y;
          } else {
		rn++;
		rsy += y;
		rsy2 += y*y;
          }
      }
   }
#pragma omp atomic
   sl.n+=ln;
#pragma omp atomic
   sr.n+=rn;
#pragma omp atomic
   sl.sy+=lsy;
#pragma omp atomic
   sr.sy+=rsy;
#pragma omp atomic
   sl.sy2+=lsy2;
#pragma omp atomic
   sr.sy2+=rsy2;
}

#pragma omp parallel private(xx,y,lnf,rnf,lsf,lsf2,rsf,rsf2)
{
   lnf=0;rnf=0;lsf=0.0,rsf=0.0,lsf2=0.0,rsf2=0.0;
#pragma omp for nowait
	for(size_t i=0;i<di.nf;i++) {
		xx = di.xf + i*di.p;
		if(nx==x.bn(xx,xi)) { //does the bottom node = xx's bottom node
			y = di.yf[i];
			if(xx[v] < xi[v][c]) {
				lnf++;
				lsf += y;
				lsf2 += y*y;
			} else {
				rnf++;
				rsf += y;
				rsf2 += y*y;
			}
		}
	}
#pragma omp atomic
   sl.nf+=lnf;
#pragma omp atomic
   sr.nf+=rnf;
#pragma omp atomic
   sl.sf+=lsf;
#pragma omp atomic
   sr.sf+=rsf;
#pragma omp atomic
   sl.sf2+=lsf2;
#pragma omp atomic
   sr.sf2+=rsf2;
}

}

// Get sufficient stats for pair of bottom children nl(left) and nr(right) in tree x for the calibart model
void getCalisuff(tree& x, tree::tree_cp nl, tree::tree_cp nr, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr)
{
   double *xx;//current x
   double y;  //current y
   sl.n=0;sl.nf=0;sl.sy=0.0;sl.sf=0.0;sl.sy2=0.0;sl.sf2=0.0;
   sr.n=0;sr.nf=0;sr.sy=0.0;sr.sf=0.0;sr.sy2=0.0;sr.sf2=0.0;

   size_t ln,rn;
   double lsy,lsy2,rsy,rsy2;
   size_t lnf,rnf;
   double lsf,lsf2,rsf,rsf2;

#pragma omp parallel private(xx,y,ln,rn,lsy,rsy,lsy2,rsy2)
{
   ln=0;rn=0;lsy=0.0,rsy=0.0,lsy2=0.0,rsy2=0.0;
#pragma omp for nowait
   for(size_t i=0;i<di.n;i++) {
      xx = di.x + i*di.p;
      tree::tree_cp bn = x.bn(xx,xi);
      if(bn==nl) {
         y = di.y[i];
	   ln++;
	   lsy += y;
	   lsy2 += y*y;
      }
      if(bn==nr) {
         y = di.y[i];
	   rn++;
	   rsy += y;
	   rsy2 += y*y;
      }
   }
#pragma omp atomic
   sl.n+=ln;
#pragma omp atomic
   sr.n+=rn;
#pragma omp atomic
   sl.sy+=lsy;
#pragma omp atomic
   sr.sy+=rsy;
#pragma omp atomic
   sl.sy2+=lsy2;
#pragma omp atomic
   sr.sy2+=rsy2;
}

#pragma omp parallel private(xx,y,lnf,rnf,lsf,rsf,lsf2,rsf2)
{
   lnf=0;rnf=0;lsf=0.0,rsf=0.0,lsf2=0.0,rsf2=0.0;
#pragma omp for nowait
   for(size_t i=0;i<di.nf;i++) {
      xx = di.xf + i*di.p;
      tree::tree_cp bn = x.bn(xx,xi);
      if(bn==nl) {
         y = di.yf[i];
	 lnf++;
	 lsf += y;
	 lsf2 += y*y;
      }
      if(bn==nr) {
         y = di.yf[i];
	 rnf++;
	 rsf += y;
	 rsf2 += y*y;
      }
   }
#pragma omp atomic
   sl.nf+=lnf;
#pragma omp atomic
   sr.nf+=rnf;
#pragma omp atomic
   sl.sf+=lsf;
#pragma omp atomic
   sr.sf+=rsf;
#pragma omp atomic
   sl.sf2+=lsf2;
#pragma omp atomic
   sr.sf2+=rsf2;
}

}

void caliallsuff(tree& x, xinfo& xi, dinfo& di, tree::npv& bnv, std::vector<sinfo>& sv)
{
	tree::tree_cp tbn; //the pointer to the bottom node for the current observations
	size_t ni;         //the index into vector of the current bottom node
	double* xx;        //current x
	double y;          //current y

	bnv.clear();
	x.getbots(bnv);

	typedef tree::npv::size_type bvsz;
	bvsz nb = bnv.size();
	sv.resize(nb);

	std::map<tree::tree_cp,size_t> bnmap;

	for(bvsz i=0;i!=bnv.size();i++) {
		bnmap[bnv[i]]=i;
		sv[i].n=0;
		sv[i].nf=0;
		sv[i].sy=0.0;
		sv[i].sf=0.0;
		sv[i].sy2=0.0;
		sv[i].sf2=0.0;
	}

#pragma omp parallel private(xx,y,tbn,ni)
{
	std::vector<size_t> n(nb,0);
	std::vector<size_t> nf(nb,0);
	std::vector<double> sy(nb,0.0);
	std::vector<double> sf(nb,0.0);
	std::vector<double> sy2(nb,0.0);
	std::vector<double> sf2(nb,0.0);

#pragma omp for nowait
	for(size_t i=0;i<di.n;i++)
	{
		xx = di.x + i*di.p;
		y = di.y[i];
		tbn = x.bn(xx,xi);
		ni = bnmap[tbn];
		++n[ni];
		sy[ni] += y;
		sy2[ni] += y*y;
	}

#pragma omp for nowait
	for(size_t i=0;i<di.nf;i++)
	{
		xx = di.xf + i*di.p;
		y = di.yf[i];
		tbn = x.bn(xx,xi);
		ni = bnmap[tbn];
		++nf[ni];
		sf[ni] += y;
		sf2[ni] += y*y;
	}

	for(bvsz i=0;i!=bnv.size();i++) {
#pragma omp atomic
		sv[i].n += n[i];
#pragma omp atomic
		sv[i].sy += sy[i];
#pragma omp atomic
		sv[i].sy2 += sy2[i];
#pragma omp atomic
		sv[i].nf += nf[i];
#pragma omp atomic
		sv[i].sf += sf[i];
#pragma omp atomic
		sv[i].sf2 += sf2[i];
	}
} // end of pragma omp parallel

}

#else  //_OPENMP
//--------------------------------------------------
// Get sufficient stats for children (v,c) of node nx in tree x for the calibart model
void getCalisuff(tree& x, tree::tree_cp nx, size_t v, size_t c, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr)
{
   double *xx;//current x
   double y;  //current y
   sl.n=0;sl.nf=0;sl.sy=0.0;sl.sf=0.0;sl.sy2=0.0;sl.sf2=0.0;
   sr.n=0;sr.nf=0;sr.sy=0.0;sr.sf=0.0;sr.sy2=0.0;sr.sf2=0.0;

   for(size_t i=0;i<di.n;i++) {
      xx = di.x + i*di.p;
      if(nx==x.bn(xx,xi)) { //does the bottom node = xx's bottom node
         y = di.y[i];
         if(xx[v] < xi[v][c]) {
               sl.n++;
               sl.sy += y;
               sl.sy2 += y*y;
          } else {
               sr.n++;
               sr.sy += y;
               sr.sy2 += y*y;
          }
      }
   }

	for(size_t i=0;i<di.nf;i++) {
		xx = di.xf + i*di.p;
		if(nx==x.bn(xx,xi)) { //does the bottom node = xx's bottom node
			y = di.yf[i];
			if(xx[v] < xi[v][c]) {
				sl.nf++;
				sl.sf += y;
				sl.sf2 += y*y;
			} else {
				sr.nf++;
				sr.sf += y;
				sr.sf2 += y*y;
			}
		}
	}
}

// Get sufficient stats for pair of bottom children nl(left) and nr(right) in tree x for the CALIBART calibration model
void getCalisuff(tree& x, tree::tree_cp nl, tree::tree_cp nr, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr)
{
   double *xx;//current x
   double y;  //current y
   sl.n=0;sl.nf=0;sl.sy=0.0;sl.sf=0.0;sl.sy2=0.0;sl.sf2=0.0;
   sr.n=0;sr.nf=0;sr.sy=0.0;sr.sf=0.0;sr.sy2=0.0;sr.sf2=0.0;

   for(size_t i=0;i<di.n;i++) {
      xx = di.x + i*di.p;
      tree::tree_cp bn = x.bn(xx,xi);
      if(bn==nl) {
         y = di.y[i];
         sl.n++;
         sl.sy += y;
         sl.sy2 += y*y;
      }
      if(bn==nr) {
         y = di.y[i];
         sr.n++;
         sr.sy += y;
         sr.sy2 += y*y;
      }
   }

   for(size_t i=0;i<di.nf;i++) {
      xx = di.xf + i*di.p;
      tree::tree_cp bn = x.bn(xx,xi);
      if(bn==nl) {
         y = di.yf[i];
         sl.nf++;
         sl.sf += y;
         sl.sf2 += y*y;
      }
      if(bn==nr) {
         y = di.yf[i];
         sr.nf++;
         sr.sf += y;
         sr.sf2 += y*y;
      }
   }
}

void caliallsuff(tree& x, xinfo& xi, dinfo& di, tree::npv& bnv, std::vector<sinfo>& sv)
{
	tree::tree_cp tbn; //the pointer to the bottom node for the current observations
	size_t ni;         //the index into vector of the current bottom node
	double* xx;        //current x
	double y;          //current y

	bnv.clear();
	x.getbots(bnv);

	typedef tree::npv::size_type bvsz;
	bvsz nb = bnv.size();
	sv.resize(nb);

	std::map<tree::tree_cp,size_t> bnmap;
	for(bvsz i=0;i!=bnv.size();i++) {
		bnmap[bnv[i]]=i;
		sv[i].n=0;
		sv[i].nf=0;
		sv[i].sy=0.0;
		sv[i].sf=0.0;
		sv[i].sy2=0.0;
		sv[i].sf2=0.0;
	}

	for(size_t i=0;i<di.n;i++)
	{
		xx = di.x + i*di.p;
		y = di.y[i];
		tbn = x.bn(xx,xi);
		ni = bnmap[tbn];
		++sv[ni].n;
		sv[ni].sy += y;
		sv[ni].sy2 += y*y;
	}
	for(size_t i=0;i<di.nf;i++)
	{
		xx = di.xf + i*di.p;
		y = di.yf[i];
		tbn = x.bn(xx,xi);
		ni = bnmap[tbn];
		++sv[ni].nf;
		sv[ni].sf += y;
		sv[ni].sf2 += y*y;
	}
}
#endif //_OPENMP

//--------------------------------------------------
// for perturb/chgv moves.
void getcalipertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, xinfo& xi, dinfo& di,
                     std::vector<sinfo>& svold, std::vector<sinfo>& svnew)
{
	caliallsuff(x,xi,di,bnv,svnew);
	pertnode->setc(oldc);
	caliallsuff(x,xi,di,bnv,svold);
}
void getcalipertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, size_t oldv, bool didswaplr, xinfo& xi, dinfo& di,
                     std::vector<sinfo>& svold, std::vector<sinfo>& svnew)
{
	caliallsuff(x,xi,di,bnv,svnew);
	pertnode->setc(oldc);
	pertnode->setv(oldv);
	if(didswaplr) //undo swap
		pertnode->swaplr();
	caliallsuff(x,xi,di,bnv,svold);
}

//--------------------------------------------------
// draw bottom node mu's for the calibart model.
void calidrmu(tree& t, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen)
{
	tree::npv bnv;
	std::vector<sinfo> sv;
	double b, bf, ybar,ybarf;
	double sig2 = pi.sigma*pi.sigma;
	double sigf2 = pi.sigmaf*pi.sigmaf;
	double a = 1.0/(pi.tau0*pi.tau0);
	double ac = 1.0/(pi.tauc*pi.tauc);
	double ad = 1.0/(pi.taud*pi.taud);

	caliallsuff(t,xi,di,bnv,sv);

	typedef tree::npv::size_type bvsz;
	for(bvsz i=0;i!=bnv.size();i++)
	{
		if(bnv[i]->isrightchildofvsplit(pi.imdim)) {
			bf = sv[i].nf/sigf2;
			ybarf = sv[i].sf/sigf2; //sv[i].nf;
			bnv[i]->setm((ybarf+pi.mud*ad)/(ad+bf) + gen.normal()/sqrt(ad+bf));
		}
		else if(bnv[i]->isleftchildofvsplit(pi.imdim)) {
			b = sv[i].n/sig2;
			ybar = sv[i].sy/sig2; //sv[i].n;
			bnv[i]->setm((ybar + pi.muc*ac)/(ac+b) + gen.normal()/sqrt(ac+b));
		}
		else {
			b = sv[i].n/sig2;
			ybar = sv[i].sy/sig2;
			bf = sv[i].nf/sigf2;
			ybarf = sv[i].sf/sigf2;
			bnv[i]->setm((ybar + ybarf + pi.mu0*a)/(a+b+bf) + gen.normal()/sqrt(a+b+bf));
		}
	}
	
}

//--------------------------------------------------
// draw the calibration parameters for the calibart model.
#ifdef _OPENMP
void calidrcp(std::vector<tree>& t, xinfo& xi, dinfo& di, pinfo& pi,
              std::vector<double> &xf, std::vector<double> &xft, std::vector<double> &yf, double *allfitf, std::vector<double> &y, double *allfit, RNG& gen)
{
	tree::npv bnv;
	std::vector<sinfo> sv;
	double loglold=0.0, loglnew=0.0;
	double theta;
	double *tempf = new double[di.nf];
	bool hardreject=false;
	double alpha,alpha2,L,U;
	typedef tree::npv::size_type bvsz;
	size_t m=t.size();

	// calculate suff stats for initial fit of the field data
#pragma omp parallel
{
	double llold=0.0;
#pragma omp for nowait
	for(size_t j=0;j<di.nf;j++)
		llold+=(yf[j]-allfitf[j])*(yf[j]-allfitf[j]);
	llold= -llold/2.0/pi.sigmaf/pi.sigmaf;
#pragma omp atomic
	loglold+= llold;
}

	for(size_t i=0;i<pi.ncaldims;i++)
	{
		loglnew=0.0;
		//propose new theta for this dimension of calibration parameter vector
		L=0.0; U = 1.0;  //we assume everything is scaled to [0,1], which the R code ensures.
		theta=pi.calests[i]+gen.normal()*pi.calpropscale[i];

		if(theta<L || theta>U)
			hardreject=true;

		xft=xf; //backup the current xf's

		//calculate the new suff stats for the proposed theta
		if(!hardreject) {
#pragma omp parallel for
			for(size_t j=0;j<di.nf;j++)
				xf[di.p*j+pi.caldims[i]]=theta;

			makefieldpred(di,xi,t,tempf);

#pragma omp parallel
{
			double llnew=0.0;
#pragma omp for nowait
			for(size_t j=0;j<di.nf;j++)
				llnew+=(yf[j]-tempf[j])*(yf[j]-tempf[j]);
			llnew= -llnew/2.0/pi.sigmaf/pi.sigmaf;
#pragma omp atomic
			loglnew+= llnew;
}
			alpha2=exp(loglnew-loglold);
			alpha=std::min(1.0,alpha2);

			for(size_t j=0;j<m;j++) {
				caliallsuff(t[j],xi,di,bnv,sv);
				for(bvsz k=0;k!=bnv.size();k++)
					if((sv[k].n+sv[k].nf)<minperbot) {
						hardreject=true;
						alpha=0.0;
						j=m; k=bnv.size()-1; //escape loop
					}
			}
		}
		else alpha=0.0;
		if(gen.uniform()<alpha) {
			pi.calests[i]=theta;
			loglold=loglnew;
			pi.calacc[i]++;
#pragma omp parallel for
			for(size_t j=0;j<di.nf;j++)
				allfitf[j] = tempf[j];
		}
		else { //reject
			xf=xft;  //swap the previous xf's back into xf.
			hardreject=false; //switch it back off in case we rejected due to a hardreject
		}
	}
	
}
#else
//_OPENMP
void calidrcp(std::vector<tree>& t, xinfo& xi, dinfo& di, pinfo& pi,
              std::vector<double> &xf, std::vector<double> &xft, std::vector<double> &yf, double *allfitf, std::vector<double> &y, double *allfit, RNG& gen)
{
	tree::npv bnv;
	std::vector<sinfo> sv;
	double loglold=0.0, loglnew=0.0;
	double theta;
	double *tempf = new double[di.nf];
	bool hardreject=false;
	double alpha,alpha2,L,U;
	typedef tree::npv::size_type bvsz;
	size_t m=t.size();



	// calculate suff stats for initial fit of the field data
	for(size_t j=0;j<di.nf;j++)
		loglold+=(yf[j]-allfitf[j])*(yf[j]-allfitf[j]);
	loglold= -loglold/2.0/pi.sigmaf/pi.sigmaf;

	for(size_t i=0;i<pi.ncaldims;i++)
	{
		loglnew=0.0;
		//propose new theta for this dimension of calibration parameter vector
		L=0.0; U = 1.0;  //we assume everything is scaled to [0,1], which the R code ensures
		theta=pi.calests[i]+gen.normal()*pi.calpropscale[i];

		if(theta<L || theta>U)
			hardreject=true;

		xft=xf; //backup the current xf's

		//calculate the new suff stats for the proposed theta
		if(!hardreject) {
			for(size_t j=0;j<di.nf;j++)
				xf[di.p*j+pi.caldims[i]]=theta;

			makefieldpred(di,xi,t,tempf);

			for(size_t j=0;j<di.nf;j++)
				loglnew+=(yf[j]-tempf[j])*(yf[j]-tempf[j]);
			loglnew= -loglnew/2.0/pi.sigmaf/pi.sigmaf;

			alpha2=exp(loglnew-loglold);
			alpha=std::min(1.0,alpha2);

			for(size_t j=0;j<m;j++) {
				caliallsuff(t[j],xi,di,bnv,sv);
				for(bvsz k=0;k!=bnv.size();k++)
					if((sv[k].n+sv[k].nf)<minperbot) {
						hardreject=true;
						alpha=0.0;
						j=m; k=bnv.size()-1; //escape loop
					}
			}
		}
		else alpha=0.0;

		if(gen.uniform()<alpha) {
			pi.calests[i]=theta;
			loglold=loglnew;
			pi.calacc[i]++;
			for(size_t j=0;j<di.nf;j++)
				allfitf[j] = tempf[j];
		}
		else { //reject
			xf=xft;  //swap the previous xf's back into xf.
			hardreject=false; //switch it back off in case we rejected due to a hardreject
		}
	}
	
}
#endif
// _OPEN_MP



#ifdef CALIBART
#ifdef _OPENMP
//---------------------------------------------------
// construct prediction of f given a single tree, overwrites existing values in ppredmean if there are any.
void makefieldpred(dinfo di, xinfo &xi, std::vector<tree> &t, double *fpred)
{
	double* fpredtemp=0; //temporary fit vector to compute prediction
	size_t m;

	fpredtemp = new double[di.nf];
	m=t.size();

#pragma omp parallel for
	for(size_t i=0;i<di.nf;i++) fpred[i]=0.0;

	for(size_t j=0;j<m;j++) {
		fitfield(t[j],xi,di,fpredtemp);
#pragma omp parallel for
		for(size_t k=0;k<di.nf;k++) fpred[k] += fpredtemp[k];
	}

	delete[] fpredtemp;
}

#else 
//_OPENMP
//---------------------------------------------------
// construct prediction of f given a single tree, overwrites existing values in ppredmean if there are any.
void makefieldpred(dinfo di, xinfo &xi, std::vector<tree> &t, double *fpred)
{
	double* fpredtemp=0; //temporary fit vector to compute prediction
	size_t m;

	fpredtemp = new double[di.nf];
	m=t.size();
	for(size_t i=0;i<di.nf;i++) fpred[i]=0.0;
	for(size_t j=0;j<m;j++) {
		fitfield(t[j],xi,di,fpredtemp);
		for(size_t k=0;k<di.nf;k++) fpred[k] += fpredtemp[k];
	}

	delete[] fpredtemp;
}
#endif
//_OPENMP
//--------------------------------------------------
// predict the simulator instead.
void makesimpred(dinfo di, xinfo& xi, std::vector<tree> &t, double *ypred)
{
	double *ypredtemp=0;
	size_t m;

	ypredtemp = new double[di.n];
	m=t.size();
	for(size_t i=0;i<di.n;i++) ypred[i]=0.0;
	for(size_t j=0;j<m;j++) {
		fit(t[j],xi,di,ypredtemp);
		for(size_t k=0;k<di.n;k++) ypred[k]+=ypredtemp[k];
	}
	
	delete[] ypredtemp;
}

#ifdef _OPENMP
//--------------------------------------------------
// fit
void fitfield(tree& t, xinfo& xi, dinfo& di, double* fv)
{
   double *xx;
   tree::tree_cp bn;

#pragma omp parallel private(xx,bn)
{
#pragma omp for
   for(size_t i=0;i<di.nf;i++) {
      xx = di.xf + i*di.p;
      bn = t.bn(xx,xi);
      fv[i] = bn->getm();
   }
}
}
#else 
//_OPENMP
//--------------------------------------------------
// fit
void fitfield(tree& t, xinfo& xi, dinfo& di, double* fv)
{
   double *xx;
   tree::tree_cp bn;
   for(size_t i=0;i<di.nf;i++) {
      xx = di.xf + i*di.p;
      bn = t.bn(xx,xi);
      fv[i] = bn->getm();
   }
}
#endif 
//_OPENMP


#endif
#endif
